from enum import Enum
from functools import lru_cache, wraps
from typing import (
    Any,
    Callable,
    Collection,
    Mapping,
    Optional,
    Sequence,
    Type,
    Union,
    cast,
    overload,
)

from apischema.aliases import Aliaser
from apischema.cache import cache
from apischema.conversions import identity
from apischema.conversions.conversions import AnyConversion, DefaultConversion
from apischema.conversions.utils import Converter
from apischema.conversions.visitor import (
    CachedConversionsVisitor,
    Serialization,
    SerializationVisitor,
    sub_conversion,
)
from apischema.fields import FIELDS_SET_ATTR, fields_set
from apischema.objects import AliasedStr, ObjectField
from apischema.objects.visitor import SerializationObjectVisitor
from apischema.serialization.serialized_methods import get_serialized_methods
from apischema.types import AnyType, NoneType, Undefined, UndefinedType
from apischema.typing import is_new_type, is_type_var, is_typed_dict
from apischema.utils import (
    Lazy,
    context_setter,
    deprecate_kwargs,
    get_origin_or_type,
    get_origin_or_type2,
    opt_or,
)
from apischema.visitor import Unsupported

NEITHER_NONE_NOR_UNDEFINED = object()
NOT_NONE = object()


SerializationMethod = Callable[[Any], Any]


def instance_checker(tp: AnyType) -> Callable[[Any], bool]:
    tp = get_origin_or_type2(tp)
    if isinstance(tp, type):
        return lambda obj: isinstance(obj, tp)
    elif is_new_type(tp):
        return instance_checker(tp.__supertype__)
    elif is_type_var(tp):
        return lambda obj: True
    else:
        raise TypeError(f"{tp} is not supported in union serialization")


class SerializationMethodVisitor(
    CachedConversionsVisitor,
    SerializationVisitor[SerializationMethod],
    SerializationObjectVisitor[SerializationMethod],
):
    def __init__(
        self,
        aliaser: Aliaser,
        fall_back_on_any: bool,
        check_type: bool,
        default_conversion: DefaultConversion,
        exclude_unset: bool,
        allow_undefined: bool,
    ):
        super().__init__(default_conversion)
        self.aliaser = aliaser
        self._fall_back_on_any = fall_back_on_any
        self._check_type = check_type
        self._exclude_unset = exclude_unset
        self._allow_undefined = allow_undefined

    def _cache_result(self, lazy: Lazy[SerializationMethod]) -> SerializationMethod:
        rec_method = None

        def method(obj: Any) -> Any:
            nonlocal rec_method
            if rec_method is None:
                rec_method = lazy()
            return rec_method(obj)

        return method

    @property
    def _any_method(self) -> Callable[[type], SerializationMethod]:
        return serialization_method_factory(
            self.aliaser,
            self._fall_back_on_any,
            self._check_type,
            self._conversions,
            self.default_conversion,
            self._exclude_unset,
            allow_undefined=self._allow_undefined,
        )

    def _wrap_type_check(
        self, cls: type, method: SerializationMethod
    ) -> SerializationMethod:
        if not self._check_type:
            return method
        fall_back_on_any, any_method = self._fall_back_on_any, self._any_method

        @wraps(method)
        def wrapper(obj: Any) -> Any:
            if isinstance(obj, cls):
                return method(obj)
            elif fall_back_on_any:
                return any_method(obj.__class__)(obj)
            else:
                raise TypeError(f"Expected {cls}, found {obj.__class__}")

        return wrapper

    def any(self) -> SerializationMethod:
        any_method = self._any_method

        def method(obj: Any) -> Any:
            return any_method(obj.__class__)(obj)

        return method

    def collection(
        self, cls: Type[Collection], value_type: AnyType
    ) -> SerializationMethod:
        serialize_value = self.visit(value_type)

        def method(obj: Any) -> Any:
            return [serialize_value(elt) for elt in obj]

        return self._wrap_type_check(cls, method)

    def enum(self, cls: Type[Enum]) -> SerializationMethod:
        any_method = self._any_method

        def method(obj: Any) -> Any:
            return any_method(obj.value.__class__)(obj.value)

        return self._wrap_type_check(cls, method)

    def literal(self, values: Sequence[Any]) -> SerializationMethod:
        return self.any()

    def mapping(
        self, cls: Type[Mapping], key_type: AnyType, value_type: AnyType
    ) -> SerializationMethod:
        serialize_key, serialize_value = self.visit(key_type), self.visit(value_type)

        def method(obj: Any) -> Any:
            return {
                serialize_key(key): serialize_value(value) for key, value in obj.items()
            }

        return self._wrap_type_check(cls, method)

    def object(self, tp: Type, fields: Sequence[ObjectField]) -> SerializationMethod:
        with context_setter(self) as setter:
            setter._allow_undefined = True
            normal_fields, aggregate_fields = [], []
            for field in fields:
                serialize_field = self.visit_with_conv(field.type, field.serialization)
                if field.is_aggregate:
                    aggregate_fields.append((field.name, serialize_field))
                else:
                    normal_fields.append(
                        (field.name, self.aliaser(field.alias), serialize_field)
                    )
            serialized_methods = [
                (
                    self.aliaser(name),
                    method.func,
                    self.visit_with_conv(types["return"], method.conversion),
                )
                for name, (method, types) in get_serialized_methods(tp).items()
            ]
            exclude_unset = self._exclude_unset

        def method(obj: Any) -> Any:
            normal_fields2, aggregate_fields2 = normal_fields, aggregate_fields
            if exclude_unset and hasattr(obj, FIELDS_SET_ATTR):
                fields_set_ = fields_set(obj)
                normal_fields2 = [
                    (name, alias, method)
                    for (name, alias, method) in normal_fields
                    if name in fields_set_
                ]
                aggregate_fields2 = [
                    (name, method)
                    for (name, method) in aggregate_fields
                    if name in fields_set_
                ]
            result = {}
            # aggregate before normal fields to avoid overloading
            for name, field_method in aggregate_fields2:
                attr = getattr(obj, name)
                result.update(field_method(attr))
            for name, alias, field_method in normal_fields2:
                attr = getattr(obj, name)
                if attr is not Undefined:
                    result[alias] = field_method(attr)
            for alias, func, method in serialized_methods:
                res = func(obj)
                if res is not Undefined:
                    result[alias] = method(res)
            return result

        cls = get_origin_or_type(tp)
        if is_typed_dict(cls):
            cls = Mapping
        return self._wrap_type_check(cls, method)

    def primitive(self, cls: Type) -> SerializationMethod:
        def method(obj: Any) -> Any:
            return obj

        return self._wrap_type_check(cls, method)

    def tuple(self, types: Sequence[AnyType]) -> SerializationMethod:
        elt_deserializers = list(map(self.visit, types))

        def method(obj: Any) -> Any:
            return [
                serialize_elt(elt) for serialize_elt, elt in zip(elt_deserializers, obj)
            ]

        if self._check_type:
            wrapped = method
            fall_back_on_any, as_list = self._fall_back_on_any, self._any_method(list)

            def method(obj: Any) -> Any:
                if len(obj) == len(elt_deserializers):
                    return wrapped(obj)
                elif fall_back_on_any:
                    return as_list(obj)
                else:
                    raise TypeError(
                        f"Expected {len(elt_deserializers)}-tuple,"
                        f" found {len(obj)}-tuple"
                    )

        return self._wrap_type_check(tuple, method)

    def union(self, alternatives: Sequence[AnyType]) -> SerializationMethod:
        method_and_checks = [
            (self.visit(alt), instance_checker(alt))
            for alt in alternatives
            if alt not in (None, UndefinedType)
        ]
        none_check = None if NoneType in alternatives else NOT_NONE
        undefined_allowed = UndefinedType in alternatives and self._allow_undefined
        fall_back_on_any, any_method = self._fall_back_on_any, self._any_method

        def method(obj: Any) -> Any:
            # Optional/Undefined optimization
            if obj is none_check:
                return obj
            error = None
            for alt_method, instance_check in method_and_checks:
                if not instance_check(obj):
                    continue
                try:
                    return alt_method(obj)
                except Exception as err:
                    error = err
            if obj is Undefined and undefined_allowed:
                return obj
            if fall_back_on_any:
                try:
                    return any_method(obj.__class__)(obj)
                except Exception as err:
                    error = err
            raise error or TypeError(
                f"Expected {Union[alternatives]}, found {obj.__class__}"
            )

        return method

    def unsupported(self, tp: AnyType) -> SerializationMethod:
        try:
            return super().unsupported(tp)
        except Unsupported:
            if self._fall_back_on_any and isinstance(tp, type):
                any_method = self._any_method
                if issubclass(tp, Mapping):

                    def method(obj: Any) -> Any:
                        return {
                            any_method(key.__class__)(key): any_method(value.__class__)(
                                value
                            )
                            for key, value in obj.items()
                        }

                    return method

                elif issubclass(tp, Collection):

                    def method(obj: Any) -> Any:
                        return [any_method(elt.__class__)(elt) for elt in obj]

                    return method

            raise

    def _visit_conversion(
        self,
        tp: AnyType,
        conversion: Serialization,
        dynamic: bool,
        next_conversion: Optional[AnyConversion],
    ) -> SerializationMethod:
        with context_setter(self) as setter:
            if conversion.fall_back_on_any is not None:
                setter._fall_back_on_any = conversion.fall_back_on_any
            if conversion.exclude_unset is not None:
                setter._exclude_unset = conversion.exclude_unset
            serialize_conv = self.visit_with_conv(
                conversion.target, sub_conversion(conversion, next_conversion)
            )

        converter = cast(Converter, conversion.converter)
        if converter is identity:
            method = serialize_conv
        else:

            def method(obj: Any) -> Any:
                return serialize_conv(converter(obj))

        return self._wrap_type_check(get_origin_or_type(tp), method)

    def visit(self, tp: AnyType) -> SerializationMethod:
        if tp == AliasedStr:
            return self._wrap_type_check(AliasedStr, self.aliaser)
        return super().visit(tp)


@cache
def serialization_method_factory(
    aliaser: Optional[Aliaser],
    fall_back_on_any: Optional[bool],
    check_type: Optional[bool],
    conversion: Optional[AnyConversion],
    default_conversion: Optional[DefaultConversion],
    exclude_unset: Optional[bool],
    allow_undefined: bool = False,
) -> Callable[[AnyType], SerializationMethod]:
    @lru_cache(serialization_method_factory.cache_info().maxsize)  # type: ignore
    def factory(tp: AnyType) -> SerializationMethod:
        from apischema import settings

        return SerializationMethodVisitor(
            opt_or(aliaser, settings.aliaser),
            opt_or(fall_back_on_any, settings.serialization.fall_back_on_any),
            opt_or(check_type, settings.serialization.check_type),
            opt_or(default_conversion, settings.serialization.default_conversion),
            opt_or(exclude_unset, settings.serialization.exclude_unset),
            allow_undefined,
        ).visit_with_conv(tp, conversion)

    return factory


def serialization_method(
    type: AnyType,
    *,
    aliaser: Aliaser = None,
    fall_back_on_any: bool = None,
    check_type: bool = None,
    conversion: AnyConversion = None,
    default_conversion: DefaultConversion = None,
    exclude_unset: bool = None,
) -> SerializationMethod:
    return serialization_method_factory(
        aliaser,
        fall_back_on_any,
        check_type,
        conversion,
        default_conversion,
        exclude_unset,
    )(type)


NO_OBJ = object()


@overload
def serialize(
    type: AnyType,
    obj: Any,
    *,
    aliaser: Aliaser = None,
    fall_back_on_any: bool = None,
    check_type: bool = None,
    conversion: AnyConversion = None,
    default_conversion: DefaultConversion = None,
    exclude_unset: bool = None,
) -> Any:
    ...


@overload
def serialize(
    obj: Any,
    *,
    aliaser: Aliaser = None,
    fall_back_on_any: bool = True,
    check_type: bool = None,
    conversion: AnyConversion = None,
    default_conversion: DefaultConversion = None,
    exclude_unset: bool = None,
) -> Any:
    ...


@deprecate_kwargs({"conversions": "conversion"})  # type: ignore
def serialize(
    type: AnyType = Any,
    obj: Any = NO_OBJ,
    *,
    aliaser: Aliaser = None,
    fall_back_on_any: bool = None,
    check_type: bool = None,
    conversion: AnyConversion = None,
    default_conversion: DefaultConversion = None,
    exclude_unset: bool = None,
) -> Any:
    # Handle overloaded signature without type
    if obj is NO_OBJ:
        type, obj = Any, type
        if fall_back_on_any is None:
            fall_back_on_any = True
    return serialization_method_factory(
        aliaser=aliaser,
        fall_back_on_any=fall_back_on_any,
        check_type=check_type,
        conversion=conversion,
        default_conversion=default_conversion,
        exclude_unset=exclude_unset,
    )(type)(obj)



import sys
import warnings
from collections import defaultdict
from dataclasses import replace
from typing import (
    Callable,
    Dict,
    List,
    Optional,
    TYPE_CHECKING,
    Type,
    TypeVar,
    Union,
    overload,
)

from apischema.conversions import LazyConversion
from apischema.conversions.conversions import (
    AnyConversion,
    ConvOrFunc,
    Conversion,
    resolve_conversion,
)
from apischema.conversions.utils import Converter, is_convertible
from apischema.types import AnyType
from apischema.typing import is_type_var
from apischema.utils import (
    MethodOrProperty,
    MethodWrapper,
    get_args2,
    get_origin_or_type,
    is_method,
    method_class,
    stop_signature_abuse,
    type_dict_wrapper,
)

if TYPE_CHECKING:
    from apischema.deserialization.coercion import Coerce


_deserializers: Dict[AnyType, List[ConvOrFunc]] = type_dict_wrapper(defaultdict(list))
_serializers: Dict[AnyType, ConvOrFunc] = type_dict_wrapper({})
Deserializer = TypeVar(
    "Deserializer", bound=Union[Callable, Conversion, staticmethod, type]
)
Serializer = TypeVar("Serializer", bound=Union[Callable, Conversion, property, type])

default_deserialization: Callable[[type], Optional[AnyConversion]]
# defaultdict.get is not hashable in 3.7
if sys.version_info < (3, 8):

    def default_deserialization(tp):
        return _deserializers.get(tp)


else:
    default_deserialization = _deserializers.get  # type: ignore


def default_serialization(tp: Type) -> Optional[AnyConversion]:
    for sub_cls in getattr(tp, "__mro__", [tp]):
        if sub_cls in _serializers:
            conversion = _serializers[sub_cls]
            if (
                sub_cls == tp
                or not isinstance(conversion, (Conversion, LazyConversion))
                or conversion.inherited in (None, True)
            ):
                return conversion
    else:
        return None


def check_converter_type(tp: AnyType) -> AnyType:
    origin = get_origin_or_type(tp)
    if not is_convertible(tp):
        raise TypeError(f"{origin} is not convertible")
    if not all(map(is_type_var, get_args2(tp))):
        raise TypeError("Generic conversion doesn't support specialization")
    return origin


def _add_deserializer(conversion: ConvOrFunc, target: AnyType):
    target = check_converter_type(target)
    if conversion not in _deserializers[target]:
        _deserializers[target].append(conversion)


class DeserializerDescriptor(MethodWrapper[staticmethod]):
    def __set_name__(self, owner, name):
        super().__set_name__(owner, name)
        method = self._method.__get__(None, object)
        resolved = resolve_conversion(method, {owner.__name__: owner})
        _add_deserializer(method, resolved.target)


@overload
def deserializer(deserializer: Deserializer) -> Deserializer:
    ...


@overload
def deserializer(
    *, lazy: Callable[[], Union[Converter, Conversion]], target: Type
) -> None:
    ...


def deserializer(
    deserializer: Deserializer = None,
    *,
    lazy: Callable[[], Union[Converter, Conversion]] = None,
    target: Type = None,
):
    if deserializer is not None:
        if isinstance(deserializer, staticmethod):
            return DeserializerDescriptor(deserializer)
        elif isinstance(deserializer, LazyConversion):
            stop_signature_abuse()
        else:
            resolved = resolve_conversion(deserializer)  # type: ignore
            _add_deserializer(deserializer, resolved.target)  # type: ignore
            return deserializer
    elif lazy is not None and target is not None:

        def replace_target():
            conversion = lazy()
            if isinstance(conversion, Conversion):
                return replace(conversion, target=target)
            else:
                return Conversion(conversion, target=target)

        _add_deserializer(LazyConversion(replace_target), target)
    else:
        stop_signature_abuse()


def _add_serializer(conversion: ConvOrFunc, source: AnyType):
    source = check_converter_type(source)
    _serializers[source] = conversion


class SerializerDescriptor(MethodWrapper[MethodOrProperty]):
    def __set_name__(self, owner, name):
        super().__set_name__(owner, name)
        _add_serializer(self._method, source=owner)


@overload
def serializer(serializer: Serializer) -> Serializer:
    ...


@overload
def serializer(
    *, lazy: Callable[[], Union[Converter, Conversion]], source: Type
) -> Callable[[Serializer], Serializer]:
    ...


def serializer(
    serializer: Serializer = None,
    *,
    lazy: Callable[[], Union[Converter, Conversion]] = None,
    source: Type = None,
):
    if serializer is not None:
        if is_method(serializer) and method_class(serializer) is None:
            return SerializerDescriptor(serializer)
        elif isinstance(serializer, LazyConversion):
            stop_signature_abuse()
        else:
            resolved = resolve_conversion(serializer)
            _add_serializer(serializer, resolved.source)
            return serializer
    elif lazy is not None and source is not None:

        def replace_source():
            conversion = lazy()
            if isinstance(conversion, Conversion):
                return replace(conversion, source=source)
            else:
                return Conversion(conversion, source=source)

        _add_serializer(LazyConversion(replace_source), source)
    else:
        stop_signature_abuse()


def reset_deserializers(cls: Type):
    _deserializers.pop(cls, ...)


def reset_serializer(cls: Type):
    _deserializers.pop(cls, ...)


class InheritedDeserializer:
    def __init__(self, method: classmethod, **kwargs):
        self.method = method
        self.kwargs = kwargs

    def __set_name__(self, owner, name):
        prev_init_subclass = owner.__init_subclass__

        def init_subclass(cls, **kwargs):
            prev_init_subclass(**kwargs)
            method = self.method.__get__(None, cls)
            deserializer(Conversion(method, target=cls, **self.kwargs))

        owner.__init_subclass__ = classmethod(init_subclass)
        init_subclass(owner)
        setattr(owner, name, self.method)


ClsMethod = TypeVar("ClsMethod")


@overload
def inherited_deserializer(method: ClsMethod) -> ClsMethod:
    ...


@overload
def inherited_deserializer(
    *,
    sub_conversions: AnyConversion = None,
    additional_properties: Optional[bool] = None,
    coercion: Optional["Coerce"] = None,
    fall_back_on_default: Optional[bool] = None,
) -> Callable[[ClsMethod], ClsMethod]:
    ...


def inherited_deserializer(method=None, **kwargs):
    warnings.warn(
        "inherited_deserializer is deprecated; __init_subclasses", DeprecationWarning
    )
    if method is None:
        return lambda func: inherited_deserializer(func, **kwargs)  # type: ignore
    if not isinstance(method, classmethod):
        raise TypeError("inherited_deserializer must be called on classmethod")
    return InheritedDeserializer(method, **kwargs)


Cls = TypeVar("Cls", bound=Type)


def as_str(cls: Cls) -> Cls:
    deserializer(Conversion(cls, source=str))
    serializer(Conversion(str, source=cls))
    return cls



__all__ = ["cache", "reset", "set_size"]
import sys
from functools import lru_cache
from typing import Any, Callable, List, TypeVar, cast

_cached: List = []

Func = TypeVar("Func", bound=Callable)


def cache(func: Func) -> Func:
    if func.__qualname__.count(".") > 1:
        raise ValueError("Cached function must be declared at module level")
    cached: Any = lru_cache()(func)
    _cached.append(cached)
    return cast(Func, cached)


def reset():
    for cached in _cached:
        cached.cache_clear()


def set_size(size: int):
    for cached in _cached:
        wrapped = cached.__wrapped__
        setattr(
            sys.modules[wrapped.__module__], wrapped.__name__, lru_cache(size)(wrapped)
        )



from functools import wraps
from typing import Any, Callable, Dict, Optional, Type, TypeVar, Union

from apischema.json_schema.types import bad_type
from apischema.types import NoneType

T = TypeVar("T")

Coercer = Callable[[Type[T], Any], T]

_bool_pairs = (
    ("0", "1"),
    ("f", "t"),
    ("n", "y"),
    ("no", "yes"),
    ("false", "true"),
    ("off", "on"),
    ("ko", "ok"),
)
STR_TO_BOOL: Dict[str, bool] = {}
for false, true in _bool_pairs:
    for s, value in ((false, False), (true, True)):
        STR_TO_BOOL[s.lower()] = value
STR_NONE_VALUES = {""}


def coerce(cls: Type[T], data: Any) -> T:
    try:
        if isinstance(data, cls):
            return data
        if data is None and cls is not NoneType:
            raise ValueError
        if cls is bool and isinstance(data, str):
            return STR_TO_BOOL[data.lower()]  # type: ignore
        if cls is NoneType and data in STR_NONE_VALUES:
            return None  # type: ignore
        if cls is list and isinstance(data, str):
            raise ValueError
        return cls(data)  # type: ignore
    except (ValueError, TypeError, KeyError):
        raise bad_type(data, cls)


_coercer: Coercer = coerce

Coerce = Union[bool, Coercer]


def wrap_coercer(coercer: Coercer) -> Coercer:
    @wraps(coercer)
    def wrapper(cls, data):
        try:
            result = coercer(cls, data)
        except AssertionError:
            raise
        except Exception:
            raise bad_type(data, cls)
        if not isinstance(result, cls):
            raise bad_type(data, cls)
        return result

    return wrapper


def get_coercer(coerce: Coerce) -> Optional[Coercer]:
    if callable(coerce):
        return wrap_coercer(coerce)
    elif coerce:
        return _coercer
    else:
        return None



__all__ = [
    "alias",
    "conversion",
    "default_as_set",
    "fall_back_on_default",
    "flattened",
    "init_var",
    "post_init",
    "properties",
    "required",
    "schema",
    "skip",
    "validators",
]

from apischema.aliases import alias
from apischema.schemas import schema
from .implem import (
    conversion,
    default_as_set,
    fall_back_on_default,
    flattened,
    init_var,
    post_init,
    properties,
    required,
    skip,
    validators,
)



import collections.abc
import inspect
import re
import sys
import warnings
from contextlib import contextmanager
from dataclasses import dataclass, is_dataclass
from enum import Enum
from functools import wraps
from types import FunctionType, MappingProxyType
from typing import (
    Any,
    Awaitable,
    Callable,
    Collection,
    Container,
    Dict,
    Generic,
    Hashable,
    Iterable,
    Iterator,
    List,
    Mapping,
    MutableMapping,
    NoReturn,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from apischema.types import (
    AnyType,
    COLLECTION_TYPES,
    MAPPING_TYPES,
    OrderedDict,
    PRIMITIVE_TYPES,
)
from apischema.typing import (
    _collect_type_vars,
    generic_mro,
    get_args,
    get_origin,
    get_type_hints,
    is_annotated,
    is_type_var,
    typing_origin,
)

try:
    from apischema.typing import Annotated
except ImportError:
    Annotated = ...  # type: ignore

PREFIX = "_apischema_"

T = TypeVar("T")
U = TypeVar("U")

Lazy = Callable[[], T]


@dataclass(frozen=True)  # dataclass enable equality check
class LazyValue(Generic[T]):
    default: T

    def __call__(self) -> T:
        return self.default


if sys.version_info <= (3, 7):  # pragma: no cover
    is_dataclass_ = is_dataclass

    def is_dataclass(obj) -> bool:
        return is_dataclass_(obj) and getattr(obj, "__origin__", None) is None


def is_hashable(obj: Any) -> bool:
    try:
        hash(obj)
    except TypeError:
        return False
    else:
        return True


def opt_or(opt: Optional[T], default: U) -> Union[T, U]:
    return opt if opt is not None else default


def to_hashable(data: Union[None, int, float, str, bool, list, dict]) -> Hashable:
    if isinstance(data, list):
        return tuple(map(to_hashable, data))
    if isinstance(data, dict):
        return tuple(sorted((to_hashable(k), to_hashable(v)) for k, v in data.items()))
    return data  # type: ignore


SNAKE_CASE_REGEX = re.compile(r"_([a-z\d])")
CAMEL_CASE_REGEX = re.compile(r"[a-z\d]([A-Z])")


def to_camel_case(s: str) -> str:
    return SNAKE_CASE_REGEX.sub(lambda m: m.group(1).upper(), s)


def to_snake_case(s: str) -> str:
    return SNAKE_CASE_REGEX.sub(lambda m: "_" + m.group(1).lower(), s)


def to_pascal_case(s: str) -> str:
    camel = to_camel_case(s)
    return camel[0].upper() + camel[1:] if camel else camel


MakeDataclassField = Union[Tuple[str, AnyType], Tuple[str, AnyType, Any]]


def merge_opts(
    func: Callable[[T, T], T]
) -> Callable[[Optional[T], Optional[T]], Optional[T]]:
    def wrapper(opt1, opt2):
        if opt1 is None:
            return opt2
        if opt2 is None:
            return opt1
        return func(opt1, opt2)

    return wrapper


K = TypeVar("K")
V = TypeVar("V")


@merge_opts
def merge_opts_mapping(m1: Mapping[K, V], m2: Mapping[K, V]) -> Mapping[K, V]:
    return {**m1, **m2}


def has_type_vars(tp: AnyType) -> bool:
    return is_type_var(tp) or bool(getattr(tp, "__parameters__", ()))


TV = AnyType  # TypeVar is not supported as a type
# 10 should be enough for all builtin types
_type_vars = [TypeVar(f"T{i}") for i in range(10)]


def get_parameters(tp: AnyType) -> Iterable[TV]:
    if hasattr(tp, "__parameters__"):
        return tp.__parameters__
    elif hasattr(tp, "__orig_bases__"):
        return _collect_type_vars(tp.__orig_bases__)
    elif is_type_var(tp):
        return (tp,)
    else:
        return _type_vars


def substitute_type_vars(tp: AnyType, substitution: Mapping[TV, AnyType]) -> AnyType:
    if is_type_var(tp):
        try:
            return substitution[tp]
        except KeyError:
            return Union[tp.__constraints__] if tp.__constraints__ else Any
    elif getattr(tp, "__parameters__", ()):
        return tp[tuple(substitution.get(p, p) for p in tp.__parameters__)]
    else:
        return tp


Func = TypeVar("Func", bound=Callable)


def typed_wraps(wrapped: Func) -> Callable[[Callable], Func]:
    return cast(Func, wraps(wrapped))


def is_subclass(tp: AnyType, base: AnyType) -> bool:
    tp, base = get_origin_or_type(tp), get_origin_or_type(base)
    return tp == base or (
        isinstance(tp, type) and isinstance(base, type) and issubclass(tp, base)
    )


def _annotated(tp: AnyType) -> AnyType:
    return get_args(tp)[0] if is_annotated(tp) else tp


def get_origin_or_type(tp: AnyType) -> AnyType:
    origin = get_origin(tp)
    return origin if origin is not None else tp


def get_origin2(tp: AnyType) -> Optional[Type]:
    return get_origin(_annotated(tp))


def get_args2(tp: AnyType) -> Tuple[AnyType, ...]:
    return get_args(_annotated(tp))


def get_origin_or_type2(tp: AnyType) -> AnyType:
    tp2 = _annotated(tp)
    origin = get_origin(tp2)
    return origin if origin is not None else tp2


def keep_annotations(tp: AnyType, annotated: AnyType) -> AnyType:
    return Annotated[(tp, *get_args(annotated)[1:])] if is_annotated(annotated) else tp


def with_parameters(tp: AnyType) -> AnyType:
    return tp[tp.__parameters__] if getattr(tp, "__parameters__", ()) else tp


def is_union_of(tp: AnyType, of: AnyType) -> bool:
    return tp == of or (get_origin_or_type2(tp) == Union and of in get_args2(tp))


MethodOrProperty = Union[Callable, property]


def _method_location(method: MethodOrProperty) -> Optional[Type]:
    if isinstance(method, property):
        method = method.fget
    while hasattr(method, "__wrapped__"):
        method = method.__wrapped__  # type: ignore
    assert isinstance(method, FunctionType)
    global_name, *class_path = method.__qualname__.split(".")[:-1]
    if global_name not in method.__globals__:
        return None
    location = method.__globals__[global_name]
    for attr in class_path:
        if hasattr(location, attr):
            location = getattr(location, attr)
        else:
            break
    return location


def is_method(method: MethodOrProperty) -> bool:
    """Return if the function is method/property declared in a class"""
    return (isinstance(method, property) and is_method(method.fget)) or (
        isinstance(method, FunctionType)
        and method.__name__ != method.__qualname__
        and isinstance(_method_location(method), (type, type(None)))
        and next(iter(inspect.signature(method).parameters), None) == "self"
    )


def method_class(method: MethodOrProperty) -> Optional[Type]:
    cls = _method_location(method)
    return cls if isinstance(cls, type) else None


METHOD_WRAPPER_ATTR = f"{PREFIX}method_wrapper"


def method_wrapper(method: MethodOrProperty, name: str = None) -> Callable:
    if isinstance(method, property):
        name = name or method.fget.__name__

        @wraps(method.fget)
        def wrapper(self):
            return getattr(self, name)

    else:
        if hasattr(method, METHOD_WRAPPER_ATTR):
            return method
        name = name or method.__name__

        @wraps(method)
        def wrapper(self, *args, **kwargs):
            return getattr(self, name)(*args, **kwargs)

    setattr(wrapper, METHOD_WRAPPER_ATTR, True)
    return wrapper


class MethodWrapper(Generic[T]):
    def __init__(self, method: T):
        self._method = method

    def getter(self, func):
        self._method = self._method.getter(func)
        return self

    def setter(self, func):
        self._method = self._method.setter(func)
        return self

    def deleter(self, func):
        self._method = self._method.deleter(func)
        return self

    def __set_name__(self, owner, name):
        setattr(owner, name, self._method)

    def __call__(self, *args, **kwargs):
        raise RuntimeError("Method __set_name__ has not been called")


def method_registerer(
    arg: Optional[Callable],
    owner: Optional[Type],
    register: Callable[[Callable, Type, str], None],
):
    def decorator(method: MethodOrProperty):
        if owner is None and is_method(method) and method_class(method) is None:

            class Descriptor(MethodWrapper[MethodOrProperty]):
                def __set_name__(self, owner, name):
                    super().__set_name__(owner, name)
                    register(method_wrapper(method), owner, name)

            return Descriptor(method)
        else:
            owner2 = owner
            if is_method(method):
                if owner2 is None:
                    owner2 = method_class(method)
                method = method_wrapper(method)
            if owner2 is None:
                try:
                    hints = get_type_hints(method)
                    owner2 = get_origin_or_type2(hints[next(iter(hints))])
                except (KeyError, StopIteration):
                    raise TypeError("First parameter of method must be typed") from None
            assert not isinstance(method, property)
            register(cast(Callable, method), owner2, method.__name__)
            return method

    return decorator if arg is None else decorator(arg)


if sys.version_info < (3, 7):
    LIST_ORIGIN = List
    SET_ORIGIN = Set
    TUPLE_ORIGIN = Tuple
    DICT_ORIGIN = Dict
else:
    LIST_ORIGIN = typing_origin(list)
    SET_ORIGIN = typing_origin(set)
    TUPLE_ORIGIN = typing_origin(tuple)
    DICT_ORIGIN = typing_origin(dict)


def replace_builtins(tp: AnyType) -> AnyType:
    origin = get_origin2(tp)
    if origin is None:
        return tp
    args = tuple(map(replace_builtins, get_args2(tp)))
    replacement: Any
    if origin in COLLECTION_TYPES:
        if issubclass(origin, collections.abc.Set):
            replacement = SET_ORIGIN
        elif issubclass(origin, tuple) and (len(args) < 2 or args[1] is not ...):
            replacement = TUPLE_ORIGIN
        else:
            replacement = LIST_ORIGIN
    elif origin in MAPPING_TYPES:
        replacement = DICT_ORIGIN
    else:
        replacement = typing_origin(origin)
    res = replacement[args] if args else replacement
    return keep_annotations(res, tp)


def sort_by_annotations_position(
    cls: Type, elts: Collection[T], key: Callable[[T], str]
) -> List[T]:
    annotations: Dict[str, Any] = OrderedDict()
    for base in reversed(cls.__mro__):
        annotations.update(getattr(base, "__annotations__", ()))
    positions = {key: i for i, key in enumerate(annotations)}
    return sorted(elts, key=lambda elt: positions.get(key(elt), len(positions)))


def stop_signature_abuse() -> NoReturn:
    raise TypeError("Stop signature abuse")


empty_dict: Mapping[str, Any] = MappingProxyType({})

ITERABLE_TYPES = (
    COLLECTION_TYPES.keys()
    | MAPPING_TYPES.keys()
    | {Iterable, collections.abc.Iterable, Container, collections.abc.Container}
)


def subtyping_substitution(
    supertype: AnyType, subtype: AnyType
) -> Tuple[Mapping[AnyType, AnyType], Mapping[AnyType, AnyType]]:
    supertype, subtype = with_parameters(supertype), with_parameters(subtype)
    supertype_to_subtype, subtype_to_supertype = {}, {}
    super_origin = get_origin_or_type2(supertype)
    for base in generic_mro(subtype):
        base_origin = get_origin_or_type2(base)
        if base_origin == super_origin or (
            base_origin in ITERABLE_TYPES and super_origin in ITERABLE_TYPES
        ):
            for base_arg, super_arg in zip(get_args(base), get_args(supertype)):
                if is_type_var(super_arg):
                    supertype_to_subtype[super_arg] = base_arg
                if is_type_var(base_arg):
                    subtype_to_supertype[base_arg] = super_arg
            break
    return supertype_to_subtype, subtype_to_supertype


def literal_values(values: Sequence[Any]) -> Sequence[Any]:
    if any(type(v) not in PRIMITIVE_TYPES and not isinstance(v, Enum) for v in values):
        raise TypeError("Only primitive types are supported for Literal/Enum")
    return [v.value if isinstance(v, Enum) else v for v in values]


awaitable_origin = get_origin(Awaitable[Any])


def is_async(func: Callable, types: Mapping[str, AnyType] = None) -> bool:
    wrapped_func = func
    while hasattr(wrapped_func, "__wrapped__"):
        wrapped_func = wrapped_func.__wrapped__  # type: ignore
    if inspect.iscoroutinefunction(wrapped_func):
        return True
    if types is None:
        try:
            types = get_type_hints(func)
        except Exception:
            types = {}
    return get_origin_or_type2(types.get("return")) == awaitable_origin


@contextmanager
def context_setter(obj: T) -> Iterator[T]:
    prev_values = {}

    class AttrSetter:
        def __getattribute__(self, name):
            return getattr(obj, name)

        def __setattr__(self, name, value):
            if name not in prev_values:
                prev_values[name] = getattr(obj, name)
            setattr(obj, name, value)

    setter = AttrSetter()
    try:
        yield cast(T, setter)
    finally:
        for attr, prev_value in prev_values.items():
            setattr(obj, attr, prev_value)


def wrap_generic_init_subclass(init_subclass: Func) -> Func:
    if sys.version_info >= (3, 7):
        return init_subclass

    @wraps(init_subclass)
    def wrapper(cls, **kwargs):
        if getattr(cls, "__origin__", None) is not None:
            return
        init_subclass(cls, **kwargs)

    return wrapper


# # Because hash of generic classes is changed by metaclass after __init_subclass__
# # classes registered in global dictionaries are no more accessible. Here is a dictionary
# # wrapper to fix this issue
if sys.version_info < (3, 7):
    K = TypeVar("K")
    V = TypeVar("V")

    class KeyWrapper:
        def __init__(self, key):
            self.key = key

        def __eq__(self, other):
            return self.key == self.key

        def __hash__(self):
            return hash(
                id(self.key)
                if getattr(self.key, "__origin__", ...) is None
                else self.key
            )

    class type_dict_wrapper(MutableMapping[K, V]):
        def __init__(self, wrapped: Dict[K, V]):
            self.wrapped = cast(Dict[KeyWrapper, V], wrapped)

        def __delitem__(self, key: K) -> None:
            del self.wrapped[KeyWrapper(key)]

        def __getitem__(self, key: K) -> V:
            return self.wrapped[KeyWrapper(key)]

        def __iter__(self) -> Iterator[K]:
            return iter(wrapper.key for wrapper in list(self.wrapped))

        def __len__(self) -> int:
            return len(self.wrapped)

        def __setitem__(self, key: K, value: V):
            self.wrapped[KeyWrapper(key)] = value


else:
    D = TypeVar("D", bound=dict)

    def type_dict_wrapper(wrapped: D) -> D:
        return wrapped


def deprecate_kwargs(
    parameters_map: Mapping[str, Optional[str]]
) -> Callable[[Func], Func]:
    def decorator(func: Func) -> Func:
        @wraps(func)
        def wrapper(*args, **kwargs):
            for param, replacement in parameters_map.items():
                if param in kwargs:
                    instead = f", use '{replacement}' instead" if replacement else ""
                    warnings.warn(
                        f"{func.__name__} parameter '{param}' is deprecated{instead}",
                        DeprecationWarning,
                    )
                    arg = kwargs.pop(param)
                    if replacement:
                        kwargs[replacement] = kwargs.get(replacement, arg)
            return func(*args, **kwargs)

        return cast(Func, wrapper)

    return decorator



import re
import warnings
from dataclasses import dataclass
from typing import Callable, Optional, Pattern, TYPE_CHECKING, Tuple, Union

from apischema.metadata.keys import (
    CONVERSION_METADATA,
    DEFAULT_AS_SET_METADATA,
    FALL_BACK_ON_DEFAULT_METADATA,
    FLATTENED_METADATA,
    INIT_VAR_METADATA,
    POST_INIT_METADATA,
    PROPERTIES_METADATA,
    REQUIRED_METADATA,
    SKIP_METADATA,
    VALIDATORS_METADATA,
)
from apischema.types import AnyType, Metadata, MetadataImplem, MetadataMixin

if TYPE_CHECKING:
    from apischema.conversions.conversions import AnyConversion
    from apischema.validation.validators import Validator


def simple_metadata(key: str) -> Metadata:
    return MetadataImplem({key: ...})


@dataclass(frozen=True)
class ConversionMetadata(MetadataMixin):
    key = CONVERSION_METADATA
    deserialization: Optional["AnyConversion"] = None
    serialization: Optional["AnyConversion"] = None


conversion = ConversionMetadata

default_as_set = simple_metadata(DEFAULT_AS_SET_METADATA)

fall_back_on_default = simple_metadata(FALL_BACK_ON_DEFAULT_METADATA)

flattened = simple_metadata(FLATTENED_METADATA)


def init_var(tp: AnyType) -> Metadata:
    return MetadataImplem({INIT_VAR_METADATA: tp})


merged = flattened

post_init = simple_metadata(POST_INIT_METADATA)


class PropertiesMetadata(dict, Metadata):  # type: ignore
    def __init__(self):
        super().__init__({PROPERTIES_METADATA: None})

    def __call__(
        self, pattern: Union[str, Pattern, "ellipsis"]  # noqa: F821
    ) -> Metadata:
        if pattern is not ...:
            pattern = re.compile(pattern)
        return MetadataImplem({PROPERTIES_METADATA: pattern})


properties = PropertiesMetadata()

required = simple_metadata(REQUIRED_METADATA)

skip = simple_metadata(SKIP_METADATA)


@dataclass(frozen=True)
class ValidatorsMetadata(MetadataMixin):
    key = VALIDATORS_METADATA
    validators: Tuple["Validator", ...]


def validators(*validator: Callable) -> ValidatorsMetadata:
    from apischema.validation.validators import Validator

    return ValidatorsMetadata(tuple(map(Validator, validator)))


def __getattr__(name):
    if name == "merged":
        warnings.warn(
            "metadata.merged is deprecated, use metadata.flattened instead",
            DeprecationWarning,
        )



from typing import Any, Mapping, Sequence, Type, TypeVar, Union, cast, overload

from apischema.cache import cache
from apischema.objects.fields import ObjectField
from apischema.objects.visitor import ObjectVisitor
from apischema.types import AnyType, OrderedDict
from apischema.typing import _GenericAlias
from apischema.visitor import Unsupported


class GetFields(ObjectVisitor[Sequence[ObjectField]]):
    def object(self, cls: Type, fields: Sequence[ObjectField]) -> Sequence[ObjectField]:
        return fields


@cache
def object_fields(tp: AnyType) -> Mapping[str, ObjectField]:
    try:
        return OrderedDict((f.name, f) for f in GetFields().visit(tp))
    except Unsupported:
        raise TypeError(f"{tp} doesn't have fields")


def object_fields2(obj: Any) -> Mapping[str, ObjectField]:
    return object_fields(
        obj if isinstance(obj, (type, _GenericAlias)) else obj.__class__
    )


T = TypeVar("T")


class FieldGetter:
    def __init__(self, obj: Any):
        self.fields = object_fields2(obj)

    def __getattribute__(self, name: str) -> ObjectField:
        try:
            return object.__getattribute__(self, "fields")[name]
        except KeyError:
            raise AttributeError(name)


@overload
def get_field(obj: Type[T]) -> T:
    ...


@overload
def get_field(obj: T) -> T:
    ...


# Overload because of Mypy issue
# https://github.com/python/mypy/issues/9003#issuecomment-667418520
def get_field(obj: Union[Type[T], T]) -> T:
    return cast(T, FieldGetter(obj))


class AliasGetter:
    def __init__(self, obj: Any):
        self.fields = object_fields2(obj)

    def __getattribute__(self, name: str) -> str:
        try:
            return object.__getattribute__(self, "fields")[name].alias
        except KeyError:
            raise AttributeError(name)


@overload
def get_alias(obj: Type[T]) -> T:
    ...


@overload
def get_alias(obj: T) -> T:
    ...


def get_alias(obj: Union[Type[T], T]) -> T:
    return cast(T, AliasGetter(obj))



import operator as op
from collections import defaultdict
from dataclasses import dataclass, field, fields
from math import gcd
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    List,
    Mapping,
    Optional,
    Pattern,
    Sequence,
    Tuple,
    TypeVar,
)

from apischema.types import Number
from apischema.utils import merge_opts, to_hashable

T = TypeVar("T")
U = TypeVar("U")

COMPARATOR_DICT: Dict[Callable, Tuple[Callable, str]] = {
    op.lt: (max, "less than %s"),
    op.le: (max, "less than or equal to %s"),
    op.gt: (min, "greater than %s"),
    op.ge: (min, "greater than or equal to %s"),
}
PREFIX_DICT: Mapping[type, str] = {
    str: "string length",
    list: "item count",
    dict: "property count",
}
Comparator = Callable[[Any, Any], Any]
Modifier = Callable[[Any], Any]
CONSTRAINT_METADATA_KEY = "constraint"


@dataclass
class ConstraintMetadata:
    alias: str
    cls: type
    comparator: Comparator
    error: str
    merge: Callable[[T, T], T]
    modifier: Optional[Modifier] = None

    @property
    def field(self) -> Any:
        return field(default=None, metadata={CONSTRAINT_METADATA_KEY: self})


def comparison(alias: str, cls: type, comparator: Callable) -> Any:
    merge, error = COMPARATOR_DICT[comparator]
    prefix = PREFIX_DICT.get(cls)  # type: ignore
    if prefix:
        error = prefix + " " + error.replace("less", "lower")
    modifier = None if cls == float else len
    return ConstraintMetadata(alias, cls, comparator, error, merge, modifier).field


def errors_func(
    checks: Collection[Tuple[Any, Comparator, Modifier, str]]
) -> Callable[[Any], Any]:
    def errors(data: Any) -> List[str]:
        return [
            err
            for attr, comp, modif, err in checks
            if comp(data if modif is None else modif(data), attr)
        ]

    return errors


def merge_mult_of(m1: Number, m2: Number) -> Number:
    if not isinstance(m1, int) and not isinstance(m2, int):
        raise TypeError("multipleOf merging is only supported with integers")
    return m1 * m2 / gcd(m1, m2)  # type: ignore


def not_match_pattern(data: str, pattern: Pattern) -> bool:
    return not pattern.match(data)


class PatternError(str):
    def __mod__(self, pattern) -> str:
        assert isinstance(pattern, Pattern)
        return f"not matching '{pattern.pattern}'"


def merge_pattern(p1: Pattern, p2: Pattern) -> Pattern:
    raise TypeError("Cannot merge patterns")


def all_unique(data: Any) -> Any:
    return len(set(map(to_hashable, data))) == len(data)


@dataclass(frozen=True)
class Constraints:
    errors_by_type: Mapping[type, Callable[[Any], Sequence[str]]] = field(
        init=False, repr=False, hash=False, compare=False
    )
    # number
    min: Optional[Number] = comparison("minimum", float, op.lt)
    max: Optional[Number] = comparison("maximum", float, op.gt)
    exc_min: Optional[Number] = comparison("exclusiveMinimum", float, op.le)
    exc_max: Optional[Number] = comparison("exclusiveMaximum", float, op.ge)
    mult_of: Optional[Number] = ConstraintMetadata(
        "multipleOf", float, op.mod, "not a multiple of %s", merge_mult_of  # type: ignore
    ).field
    # string
    min_len: Optional[int] = comparison("minLength", str, op.lt)
    max_len: Optional[int] = comparison("maxLength", str, op.gt)
    pattern: Optional[Pattern] = ConstraintMetadata(
        "pattern", str, not_match_pattern, PatternError(), merge_pattern  # type: ignore
    ).field
    # array
    min_items: Optional[int] = comparison("minItems", list, op.lt)
    max_items: Optional[int] = comparison("maxItems", list, op.gt)
    unique: Optional[bool] = ConstraintMetadata(
        "uniqueItems", list, op.ne, "duplicate items", op.or_, modifier=all_unique
    ).field
    # object
    min_props: Optional[int] = comparison("minProperties", dict, op.lt)
    max_props: Optional[int] = comparison("maxProperties", dict, op.gt)

    @property
    def attr_and_metata(
        self,
    ) -> Collection[Tuple[str, Optional[Any], ConstraintMetadata]]:
        return [
            (f.name, getattr(self, f.name), f.metadata[CONSTRAINT_METADATA_KEY])
            for f in fields(self)
            if CONSTRAINT_METADATA_KEY in f.metadata
        ]

    def __post_init__(self):
        checks_by_type = defaultdict(list)
        for _, attr, metadata in self.attr_and_metata:
            if attr is None:
                continue
            error = metadata.error if isinstance(attr, bool) else metadata.error % attr
            error += f" ({metadata.alias})"
            checks_by_type[metadata.cls].append(
                (attr, metadata.comparator, metadata.modifier, error)
            )

        errors = {cls: errors_func(checks) for cls, checks in checks_by_type.items()}
        if float in errors:
            errors[int] = errors[float]
        object.__setattr__(self, "errors_by_type", errors)

    def merge_into(self, base_schema: Dict[str, Any]):
        for name, attr, metadata in self.attr_and_metata:
            if attr is not None:
                alias = metadata.alias
                if alias in base_schema:
                    base_schema[alias] = metadata.merge(attr, base_schema[alias])  # type: ignore
                else:
                    base_schema[alias] = attr


@merge_opts
def merge_constraints(c1: Constraints, c2: Constraints) -> Constraints:
    constraints: Dict[str, Any] = {}
    for name, attr1, metadata in c1.attr_and_metata:
        attr2 = getattr(c2, name)
        if attr1 is None:
            constraints[name] = attr2
        elif attr2 is None:
            constraints[name] = attr1
        else:
            constraints[name] = metadata.merge(attr1, attr2)  # type: ignore
    return Constraints(**constraints)  # type: ignore



from setuptools import find_packages, setup

with open("README.md") as f:
    README = f.read()

setup(
    name="apischema",
    version="0.15.0",
    url="https://github.com/wyfo/apischema",
    author="Joseph Perez",
    author_email="joperez@hotmail.fr",
    license="MIT",
    packages=find_packages(include=["apischema*"]),
    package_data={"apischema": ["py.typed"]},
    description="JSON (de)serialization + GraphQL and JSON schema generation through python typing, with a spoonful of sugar.",
    long_description=README,
    long_description_content_type="text/markdown",
    python_requires=">=3.6",
    install_requires=["dataclasses==0.7;python_version<'3.7'"],
    extras_require={
        "graphql": ["graphql-core>=3.1.2"],
        "examples": [
            "graphql-core>=3.1.2",
            "pytest>=6.2.2",
            "pydantic>=1.7.3",
            "attrs>=20.3.0",
            "sqlalchemy>=1.3.23",
            "bson>=0.5.10",
        ],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



from collections import defaultdict
from dataclasses import dataclass, replace
from enum import Enum
from functools import wraps
from typing import (
    AbstractSet,
    Any,
    Callable,
    Dict,
    Hashable,
    Iterable,
    List,
    Mapping,
    Optional,
    Pattern,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

from apischema.aliases import Aliaser
from apischema.cache import cache
from apischema.conversions.conversions import AnyConversion, DefaultConversion
from apischema.conversions.utils import identity
from apischema.conversions.visitor import (
    CachedConversionsVisitor,
    Deserialization,
    DeserializationVisitor,
    sub_conversion,
)
from apischema.dependencies import get_dependent_required
from apischema.deserialization.coercion import Coerce, Coercer, get_coercer
from apischema.deserialization.flattened import get_deserialization_flattened_aliases
from apischema.json_schema.patterns import infer_pattern
from apischema.json_schema.types import bad_type
from apischema.metadata.implem import ValidatorsMetadata
from apischema.metadata.keys import SCHEMA_METADATA, VALIDATORS_METADATA
from apischema.objects import ObjectField
from apischema.objects.fields import FieldKind
from apischema.objects.visitor import DeserializationObjectVisitor
from apischema.schemas import Schema, get_schema
from apischema.schemas.constraints import Constraints, merge_constraints
from apischema.types import (
    AnyType,
    COLLECTION_TYPES,
    MAPPING_TYPES,
    NoneType,
    OrderedDict,
    UndefinedType,
)
from apischema.typing import get_args, get_origin
from apischema.utils import (
    Lazy,
    PREFIX,
    context_setter,
    deprecate_kwargs,
    get_origin_or_type,
    literal_values,
    opt_or,
)
from apischema.validation import get_validators
from apischema.validation.errors import ErrorKey, ValidationError, merge_errors
from apischema.validation.mock import ValidatorMock
from apischema.validation.validators import Validator, validate
from apischema.visitor import Unsupported

DICT_TYPE = get_origin(Dict[Any, Any])
LIST_TYPE = get_origin(List[Any])

MISSING_PROPERTY = ValidationError(["missing property"])
UNEXPECTED_PROPERTY = ValidationError(["unexpected property"])

NOT_NONE = object()

INIT_VARS_ATTR = f"{PREFIX}_init_vars"

T = TypeVar("T")


DeserializationMethod = Callable[[Any], T]


@dataclass(frozen=True)
class DeserializationMethodFactory:
    factory: Callable[
        [Optional[Constraints], Sequence[Validator]], DeserializationMethod
    ]
    data_type: Optional[Type] = None
    coercer: Optional[Coercer] = None
    constraints: Optional[Constraints] = None
    validators: Sequence[Validator] = ()

    def merge(
        self, constraints: Optional[Constraints], validators: Sequence[Validator]
    ) -> "DeserializationMethodFactory":
        if constraints is None and not validators:
            return self
        return replace(
            self,
            constraints=merge_constraints(self.constraints, constraints),
            validators=(*validators, *self.validators),
        )

    @property
    def method(self) -> DeserializationMethod:
        method = self.factory(self.constraints, self.validators)  # type: ignore

        if self.data_type is not None and self.coercer is not None:
            wrapped_for_coercion, coercer, cls = method, self.coercer, self.data_type

            @wraps(method)
            def method(data: Any) -> Any:
                return wrapped_for_coercion(coercer(cls, data))

        return method


FallBakOnDefault = bool
Required = Union[bool, AbstractSet[str]]


def get_constraints(schema: Optional[Schema]) -> Optional[Constraints]:
    return schema.constraints if schema is not None else None


def with_validators(
    method: DeserializationMethod, validators: Sequence[Validator]
) -> DeserializationMethod:
    if not validators:
        return method

    @wraps(method)
    def wrapper(data: Any) -> Any:
        return validate(method(data), validators)

    return wrapper


def get_constraint_errors(
    constraints: Optional[Constraints], cls: type
) -> Optional[Callable[[Any], Sequence[str]]]:
    return None if constraints is None else constraints.errors_by_type.get(cls)


class DeserializationMethodVisitor(
    CachedConversionsVisitor[Deserialization, DeserializationMethodFactory],
    DeserializationVisitor[DeserializationMethodFactory],
    DeserializationObjectVisitor[DeserializationMethodFactory],
):
    def __init__(
        self,
        additional_properties: bool,
        aliaser: Aliaser,
        coercer: Optional[Coercer],
        default_conversion: DefaultConversion,
        fall_back_on_default: bool,
    ):
        super().__init__(default_conversion)
        self._additional_properties = additional_properties
        self.aliaser = aliaser
        self._coercer = coercer
        self._fall_back_on_default = fall_back_on_default

    def _cache_key(self) -> Hashable:
        return self._coercer

    def _cache_result(
        self, lazy: Lazy[DeserializationMethodFactory]
    ) -> DeserializationMethodFactory:
        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            rec_method = None

            def method(data: Any) -> Any:
                nonlocal rec_method
                if rec_method is None:
                    rec_method = lazy().merge(constraints, validators).method
                return rec_method(data)

            return method

        return DeserializationMethodFactory(factory)

    def annotated(
        self, tp: AnyType, annotations: Sequence[Any]
    ) -> DeserializationMethodFactory:
        factory = super().annotated(tp, annotations)
        for annotation in reversed(annotations):
            if isinstance(annotation, Mapping):
                factory = factory.merge(
                    get_constraints(annotation.get(SCHEMA_METADATA)),
                    annotation.get(
                        VALIDATORS_METADATA, ValidatorsMetadata(())
                    ).validators,
                )
        return factory

    def any(self) -> DeserializationMethodFactory:
        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            def method(data: Any) -> Any:
                if constraints is not None:
                    cls = type(data)
                    if cls in constraints.errors_by_type:
                        errors = constraints.errors_by_type[cls](data)
                        if errors:
                            raise ValidationError(errors)
                return data

            return with_validators(method, validators)

        return DeserializationMethodFactory(factory)

    def collection(
        self, cls: Type[Iterable], value_type: AnyType
    ) -> DeserializationMethodFactory:
        value_factory = self.visit(value_type)

        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            deserialize_value = value_factory.method
            constraint_errors = get_constraint_errors(constraints, list)

            def method(data: Any) -> Any:
                if not isinstance(data, list):
                    raise bad_type(data, list)
                elts = []
                elt_errors: Dict[ErrorKey, ValidationError] = {}
                for i, elt in enumerate(data):
                    try:
                        elts.append(deserialize_value(elt))
                    except ValidationError as err:
                        elt_errors[i] = err
                errors = constraint_errors(data) if constraint_errors else ()
                if elt_errors or errors:
                    raise ValidationError(errors, elt_errors)
                return elts if cls is LIST_TYPE else COLLECTION_TYPES[cls](elts)

            return with_validators(method, validators)

        return DeserializationMethodFactory(factory, list)

    def enum(self, cls: Type[Enum]) -> DeserializationMethodFactory:
        literal_factory = self.literal([elt.value for elt in cls])

        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            deserialize_literal = literal_factory.merge(constraints, validators).method

            def method(data: Any) -> Any:
                return cls(deserialize_literal(data))

            return method

        return DeserializationMethodFactory(factory)

    def literal(self, values: Sequence[Any]) -> DeserializationMethodFactory:
        primitive_values = literal_values(values)
        any_factory = self.any()

        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            deserialize_any = any_factory.merge(constraints, validators).method

            def method(data: Any) -> Any:
                try:
                    result = values[primitive_values.index(data)]
                except ValueError:
                    raise ValidationError([f"not one of {primitive_values}"])
                deserialize_any(data)  # for validation
                return result

            return method

        return DeserializationMethodFactory(factory)

    def mapping(
        self, cls: Type[Mapping], key_type: AnyType, value_type: AnyType
    ) -> DeserializationMethodFactory:
        key_factory, value_factory = self.visit(key_type), self.visit(value_type)

        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            deserialize_key = key_factory.method
            deserialize_value = value_factory.method
            constraint_errors = get_constraint_errors(constraints, dict)

            def method(data: Any) -> Any:
                if not isinstance(data, dict):
                    raise bad_type(data, dict)
                items = {}
                item_errors: Dict[ErrorKey, ValidationError] = {}
                for key, value in data.items():
                    assert isinstance(key, str)
                    try:
                        items[deserialize_key(key)] = deserialize_value(value)
                    except ValidationError as err:
                        item_errors[key] = err
                errors = constraint_errors(data) if constraint_errors else ()
                if item_errors or errors:
                    raise ValidationError(errors, item_errors)
                return items if cls is DICT_TYPE else MAPPING_TYPES[cls](items)

            return with_validators(method, validators)

        return DeserializationMethodFactory(factory, dict)

    def object(
        self, tp: Type, fields: Sequence[ObjectField]
    ) -> DeserializationMethodFactory:
        field_factories = [
            self.visit_with_conv(f.type, f.deserialization).merge(
                get_constraints(f.schema), f.validators
            )
            for f in fields
        ]
        additional_properties = self._additional_properties
        fall_back_on_default = self._fall_back_on_default

        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            cls = get_origin_or_type(tp)
            normal_fields: List[
                Tuple[str, str, DeserializationMethod, Required, FallBakOnDefault]
            ] = []
            flattened_fields: List[
                Tuple[str, AbstractSet[str], DeserializationMethod, FallBakOnDefault]
            ] = []
            pattern_fields: List[
                Tuple[str, Pattern, DeserializationMethod, FallBakOnDefault]
            ] = []
            additional_field: Optional[
                Tuple[str, DeserializationMethod, FallBakOnDefault]
            ] = None
            post_init_modified = {field.name for field in fields if field.post_init}
            alias_by_name = {field.name: self.aliaser(field.alias) for field in fields}
            requiring: Dict[str, Set[str]] = defaultdict(set)
            for f, reqs in get_dependent_required(cls).items():
                for req in reqs:
                    requiring[req].add(alias_by_name[f])
            init_defaults = [
                (f.name, f.default_factory)
                for f in fields
                if f.kind == FieldKind.WRITE_ONLY
            ]
            for field, field_factory in zip(fields, field_factories):
                deserialize_field = field_factory.method
                field_fall_back_on_default = (
                    field.fall_back_on_default or fall_back_on_default
                )
                if field.flattened:
                    flattened_aliases = get_deserialization_flattened_aliases(
                        cls, field, self.default_conversion
                    )
                    flattened_fields.append(
                        (
                            field.name,
                            set(map(self.aliaser, flattened_aliases)),
                            deserialize_field,
                            field_fall_back_on_default,
                        )
                    )
                elif field.pattern_properties is ...:
                    pattern_fields.append(
                        (
                            field.name,
                            infer_pattern(field.type, self.default_conversion),
                            deserialize_field,
                            field_fall_back_on_default,
                        )
                    )
                elif field.pattern_properties is not None:
                    assert isinstance(field.pattern_properties, Pattern)
                    pattern_fields.append(
                        (
                            field.name,
                            field.pattern_properties,
                            deserialize_field,
                            field_fall_back_on_default,
                        )
                    )
                elif field.additional_properties:
                    additional_field = (
                        field.name,
                        deserialize_field,
                        field_fall_back_on_default,
                    )
                else:
                    normal_fields.append(
                        (
                            field.name,
                            self.aliaser(field.alias),
                            deserialize_field,
                            field.required or requiring[field.name],
                            field_fall_back_on_default,
                        )
                    )
            has_aggregate_field = (
                flattened_fields or pattern_fields or (additional_field is not None)
            )
            constraint_errors = get_constraint_errors(constraints, dict)

            def method(data: Any) -> Any:
                if not isinstance(data, dict):
                    raise bad_type(data, dict)
                values: Dict[str, Any] = {}
                aliases: List[str] = []
                errors = list(constraint_errors(data)) if constraint_errors else []
                field_errors: Dict[ErrorKey, ValidationError] = OrderedDict()
                for (
                    name,
                    alias,
                    field_method,
                    required,
                    fall_back_on_default,
                ) in normal_fields:
                    if alias in data:
                        aliases.append(alias)
                        try:
                            values[name] = field_method(data[alias])
                        except ValidationError as err:
                            if not fall_back_on_default:
                                field_errors[alias] = err
                    elif not required:
                        pass
                    elif required is True:
                        field_errors[alias] = MISSING_PROPERTY
                    else:
                        assert isinstance(required, AbstractSet)
                        requiring = required & data.keys()
                        if requiring:
                            msg = f"missing property (required by {sorted(requiring)})"
                            field_errors[alias] = ValidationError([msg])
                if has_aggregate_field:
                    for (
                        name,
                        flattened_alias,
                        field_method,
                        fall_back_on_default,
                    ) in flattened_fields:

                        flattened = {
                            alias: data[alias]
                            for alias in flattened_alias
                            if alias in data
                        }
                        aliases.extend(flattened)
                        try:
                            values[name] = field_method(flattened)
                        except ValidationError as err:
                            if not fall_back_on_default:
                                errors.extend(err.messages)
                                field_errors.update(err.children)
                    if len(data) != len(aliases):
                        remain = data.keys() - set(aliases)
                    else:
                        remain = set()
                    for (
                        name,
                        pattern,
                        field_method,
                        fall_back_on_default,
                    ) in pattern_fields:
                        matched = {
                            key: data[key] for key in remain if pattern.match(key)
                        }
                        remain -= matched.keys()
                        try:
                            values[name] = field_method(matched)
                        except ValidationError as err:
                            if not fall_back_on_default:
                                errors.extend(err.messages)
                                field_errors.update(err.children)
                    if additional_field is not None:
                        name, field_method, fall_back_on_default = additional_field
                        additional = {key: data[key] for key in remain}
                        try:
                            values[name] = field_method(additional)
                        except ValidationError as err:
                            if not fall_back_on_default:
                                errors.extend(err.messages)
                                field_errors.update(err.children)
                    elif remain and not additional_properties:
                        for key in remain:
                            field_errors[key] = UNEXPECTED_PROPERTY
                elif len(data) != len(aliases) and not additional_properties:
                    for key in data.keys() - set(aliases):
                        field_errors[key] = UNEXPECTED_PROPERTY

                validators2: Sequence[Validator]
                if validators:
                    init: Dict[str, Any] = {}
                    for name, default_factory in init_defaults:
                        if name in values:
                            init[name] = values[name]
                        elif name not in field_errors:
                            assert default_factory is not None
                            init[name] = default_factory()
                    # Don't keep validators when all dependencies are default
                    validators2 = [
                        v for v in validators if v.dependencies & values.keys()
                    ]
                    if field_errors or errors:
                        error = ValidationError(errors, field_errors)
                        invalid_fields = field_errors.keys() | post_init_modified
                        validators2 = [
                            v
                            for v in validators2
                            if not v.dependencies & invalid_fields
                        ]
                        try:
                            validate(ValidatorMock(cls, values), validators2, **init)
                        except ValidationError as err:
                            error = merge_errors(error, err)
                        raise error
                elif field_errors or errors:
                    raise ValidationError(errors, field_errors)
                else:
                    validators2, init = (), ...  # type: ignore # only for linter
                try:
                    res = cls(**values)
                except (AssertionError, ValidationError):
                    raise
                except TypeError as err:
                    if str(err).startswith("__init__() got"):
                        raise Unsupported(cls)
                    else:
                        raise ValidationError([str(err)])
                except Exception as err:
                    raise ValidationError([str(err)])
                if validators2:
                    validate(res, validators2, **init)
                return res

            return method

        return DeserializationMethodFactory(factory, dict)

    def primitive(self, cls: Type) -> DeserializationMethodFactory:
        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            if constraints is not None and cls in constraints.errors_by_type:
                constraint_errors = constraints.errors_by_type[cls]

                def method(data: Any) -> Any:
                    if not isinstance(data, cls) and not (
                        cls == float and isinstance(data, int)
                    ):
                        raise bad_type(data, cls)
                    if data is not None:
                        errors = constraint_errors(data)
                        if errors:
                            raise ValidationError(errors)
                    return data

            else:

                def method(data: Any) -> Any:
                    if not isinstance(data, cls) and not (
                        cls == float and isinstance(data, int)
                    ):
                        raise bad_type(data, cls)
                    return data

            return with_validators(method, validators)

        return DeserializationMethodFactory(factory, cls)

    def subprimitive(self, cls: Type, superclass: Type) -> DeserializationMethodFactory:
        primitive_factory = self.primitive(superclass)

        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            deserialize_primitive = primitive_factory.merge(
                constraints, validators
            ).method

            def method(data: Any) -> Any:
                return superclass(deserialize_primitive(data))

            return method

        return DeserializationMethodFactory(factory)

    def tuple(self, types: Sequence[AnyType]) -> DeserializationMethodFactory:
        nb_elts, elt_factories = len(types), [self.visit(tp) for tp in types]

        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            elt_deserializers = [elt_factory.method for elt_factory in elt_factories]
            tuple_constraints = merge_constraints(
                constraints, Constraints(min_items=nb_elts, max_items=nb_elts)
            )
            constraint_errors = get_constraint_errors(tuple_constraints, list)

            def method(data: Any) -> Any:
                if not isinstance(data, list):
                    raise bad_type(data, list)
                elts: List[Any] = []
                elt_errors: Dict[ErrorKey, ValidationError] = {}
                for i, (deserialize_elt, elt) in enumerate(
                    zip(elt_deserializers, data)
                ):
                    try:
                        elts.append(deserialize_elt(elt))
                    except ValidationError as err:
                        elt_errors[i] = err
                errors = constraint_errors(data) if constraint_errors else ()
                if elt_errors or errors:
                    raise ValidationError(errors, elt_errors)
                return tuple(elts)

            return with_validators(method, validators)

        return DeserializationMethodFactory(factory, list)

    def union(self, alternatives: Sequence[AnyType]) -> DeserializationMethodFactory:
        none_check = None if NoneType in alternatives else NOT_NONE
        alt_factories = [
            self.visit(alt) for alt in alternatives if alt not in (None, UndefinedType)
        ]

        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            alt_deserializers = [
                factory.merge(constraints, validators).method
                for factory in alt_factories
            ]

            def method(data: Any) -> Any:
                # Optional optimization
                if data is none_check:
                    return None
                error: Optional[ValidationError] = None
                for deserialize_alt in alt_deserializers:
                    try:
                        return deserialize_alt(data)
                    except ValidationError as err:
                        error = merge_errors(error, err)
                if none_check is None:
                    error = merge_errors(error, bad_type(data, NoneType))
                if error is None:  # empty union
                    return data
                else:
                    raise error

            return method

        return DeserializationMethodFactory(factory)

    def _visit_conversion(
        self,
        tp: AnyType,
        conversion: Deserialization,
        dynamic: bool,
        next_conversion: Optional[AnyConversion],
    ) -> DeserializationMethodFactory:
        assert conversion
        conv_factories = []
        for conv in conversion:
            with context_setter(self) as setter:
                if conv.additional_properties is not None:
                    setter._additional_properties = conv.additional_properties
                if conv.fall_back_on_default is not None:
                    setter._fall_back_on_default = conv.fall_back_on_default
                if conv.coerce is not None:
                    setter._coercer = get_coercer(conv.coerce)
                sub_conv = sub_conversion(conv, next_conversion)
                conv_factories.append(self.visit_with_conv(conv.source, sub_conv))

        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            conv_factories2 = conv_factories
            if not dynamic:
                conv_factories2 = [
                    fact.merge(constraints, validators) for fact in conv_factories
                ]
            conv_deserializers = [
                (fact.method, conv.converter)
                for conv, fact in zip(conversion, conv_factories2)
            ]
            method: DeserializationMethod
            if len(conv_deserializers) > 1:

                def method(data: Any) -> Any:
                    error: Optional[ValidationError] = None
                    for deserialize_conv, converter in conv_deserializers:
                        try:
                            value = deserialize_conv(data)
                            break
                        except ValidationError as err:
                            error = merge_errors(error, err)
                    else:
                        assert error is not None
                        raise error
                    try:
                        return converter(value)  # type: ignore
                    except (ValidationError, AssertionError):
                        raise
                    except Exception as err:
                        raise ValidationError([str(err)])

            elif conv_deserializers[0][1] is identity:
                method, _ = conv_deserializers[0]
            else:
                conv_deserializer, converter = conv_deserializers[0]

                def method(data: Any) -> Any:
                    try:
                        return converter(conv_deserializer(data))  # type: ignore
                    except (ValidationError, AssertionError):
                        raise
                    except Exception as err:
                        raise ValidationError([str(err)])

            return method

        return DeserializationMethodFactory(factory)

    def visit_conversion(
        self,
        tp: AnyType,
        conversion: Optional[Deserialization],
        dynamic: bool,
        next_conversion: Optional[AnyConversion] = None,
    ) -> DeserializationMethodFactory:
        factory = super().visit_conversion(tp, conversion, dynamic, next_conversion)
        if factory.coercer is None and self._coercer is not None:
            factory = replace(factory, coercer=self._coercer)
        if not dynamic:
            factory = factory.merge(get_constraints(get_schema(tp)), get_validators(tp))
            if get_args(tp):
                factory = factory.merge(
                    get_constraints(get_schema(get_origin(tp))),
                    get_validators(get_origin(tp)),
                )
        return factory


@overload
def deserialization_method(
    type: Type[T],
    *,
    additional_properties: bool = None,
    aliaser: Aliaser = None,
    coerce: Coerce = None,
    conversion: AnyConversion = None,
    default_conversion: DefaultConversion = None,
    fall_back_on_default: bool = None,
    schema: Schema = None,
) -> DeserializationMethod[T]:
    ...


@overload
def deserialization_method(
    type: AnyType,
    *,
    additional_properties: bool = None,
    aliaser: Aliaser = None,
    coerce: Coerce = None,
    conversion: AnyConversion = None,
    default_conversion: DefaultConversion = None,
    fall_back_on_default: bool = None,
    schema: Schema = None,
) -> DeserializationMethod:
    ...


@cache
def deserialization_method(
    type: AnyType,
    *,
    additional_properties: bool = None,
    aliaser: Aliaser = None,
    coerce: Coerce = None,
    conversion: AnyConversion = None,
    default_conversion: DefaultConversion = None,
    fall_back_on_default: bool = None,
    schema: Schema = None,
) -> DeserializationMethod:
    from apischema import settings

    return (
        DeserializationMethodVisitor(
            opt_or(
                additional_properties, settings.deserialization.additional_properties
            ),
            opt_or(aliaser, settings.aliaser),
            get_coercer(opt_or(coerce, settings.deserialization.coerce)),
            opt_or(default_conversion, settings.deserialization.default_conversion),
            opt_or(fall_back_on_default, settings.deserialization.fall_back_on_default),
        )
        .visit_with_conv(type, conversion)
        .merge(get_constraints(schema), ())
        .method
    )


@overload
def deserialize(
    type: Type[T],
    data: Any,
    *,
    additional_properties: bool = None,
    aliaser: Aliaser = None,
    coerce: Coerce = None,
    conversion: AnyConversion = None,
    default_conversion: DefaultConversion = None,
    fall_back_on_default: bool = None,
    schema: Schema = None,
) -> T:
    ...


@overload
def deserialize(
    type: AnyType,
    data: Any,
    *,
    additional_properties: bool = None,
    aliaser: Aliaser = None,
    coerce: Coerce = None,
    conversion: AnyConversion = None,
    default_conversion: DefaultConversion = None,
    fall_back_on_default: bool = None,
    schema: Schema = None,
) -> Any:
    ...


@deprecate_kwargs(
    {
        "coercion": "coerce",
        "conversions": "conversion",
        "default_fallback": "fall_back_on_default",
    }
)
def deserialize(
    type: AnyType,
    data: Any,
    *,
    additional_properties: bool = None,
    aliaser: Aliaser = None,
    coerce: Coerce = None,
    conversion: AnyConversion = None,
    default_conversion: DefaultConversion = None,
    fall_back_on_default: bool = None,
    schema: Schema = None,
) -> Any:
    return deserialization_method(
        type,
        additional_properties=additional_properties,
        aliaser=aliaser,
        coerce=coerce,
        conversion=conversion,
        default_conversion=default_conversion,
        fall_back_on_default=fall_back_on_default,
        schema=schema,
    )(data)
